
#include <allegro.h>
#include <math.h>
#include "../melee.h"

// Define global space frame
SpaceFrame *spaceFrame;

SpaceLocation::SpaceLocation(double lx, double ly, double langle, int lcflag) :
  id(SPACE_LOCATION),
  x(lx),
  y(ly),
  angle(langle),
  vx(0.0),
  vy(0.0),
  driftvx(0.0),
  driftvy(0.0),
  state(TRUE),
  damage_factor(0),
  collide_flag(lcflag),
  ally_flag(lcflag),
  damage(0),
  directDamage(0),
  repair(0),
  fuel_sap(0),
  speed_loss(0.0)
{
}

inline double SpaceLocation::get_angle()
{
  return(normalize(angle, 360.0));
}

inline double SpaceLocation::get_vx()
{
  return(vx + driftvx);
}

inline double SpaceLocation::get_vy()
{
  return(vy + driftvy);
}

inline int SpaceLocation::getID()
{
  return id;
}

inline int SpaceLocation::collideFlag()
{
  return collide_flag;
}

inline int SpaceLocation::allyFlag()
{
  return ally_flag;
}

inline double SpaceLocation::normal_x()
{
  return(normalize(x, X_MAX));
}

inline double SpaceLocation::normal_y()
{
  return(normalize(y, Y_MAX));
}

inline double SpaceLocation::nearest_x(SpaceLocation *l)
{
  return(nearest_coord(normal_x(), l->normal_x(), X_MAX));
}

inline double SpaceLocation::nearest_y(SpaceLocation *l)
{
  return(nearest_coord(normal_y(), l->normal_y(), Y_MAX));
}

inline double SpaceLocation::distance(SpaceLocation *l)
{
  return(distance_from(normal_x(), normal_y(), l->normal_x(), l->normal_y()));
}

double SpaceLocation::trajectory_angle(SpaceLocation *l)
{
  double dx = min_delta(normal_x(), l->normal_x(), X_MAX);
  double dy = min_delta(normal_y(), l->normal_y(), Y_MAX);

  double alpha = atan(dy / dx);

  if(dx < 0.0)
    alpha += PI;
  alpha = normalize(alpha, PI * 2.0);

  alpha = ((alpha + PI) / (PI * 2.0)) * 360.0;
  alpha = normalize(alpha, 360.0);

  return(alpha);
}

inline int SpaceLocation::exists()
{
  return state;
}

inline int SpaceLocation::canCollide(SpaceLocation *other)
{
  if(other == NULL)
    return(FALSE);

  if((collide_flag < 0) || (other->collide_flag < 0))
    return(FALSE);

  if((collide_flag != 0) && (other->collide_flag != 0) &&
    (collide_flag == other->collide_flag))
    return(FALSE);

  return(TRUE);
}

int SpaceLocation::isAlly(SpaceLocation *other)
{
  return(ally_flag == other->ally_flag);
}

SpaceObject::SpaceObject(double ox, double oy, double oangle, int ocflag,
  SpaceSprite *osprite) :
  SpaceLocation(ox, oy, oangle, ocflag),
  w(osprite->width()),
  h(osprite->height()),
  mass(0.0),
  sprite(osprite),
  sprite_index(0),
  radar_color(-1),
  next(NULL)
{
  id = SPACE_OBJECT;
}

inline double SpaceObject::corner_x()
{
  return(wcorner(x, w));
}

inline double SpaceObject::corner_y()
{
  return(hcorner(y, h));
}

inline int SpaceObject::getSpriteIndex()
{
  return(sprite_index);
}

//MYCODE
inline SpaceSprite* SpaceObject::getSprite()
{
  return( sprite );
}

inline double SpaceObject::width()
{
  return(sprite->width());
}

inline double SpaceObject::height()
{
  return(sprite->height());
}

void SpaceObject::animate(Frame *space, Frame *radar)
{
  // animate object sprite
  sprite->animate((int)(corner_x()), (int)(corner_y()), sprite_index, space);

  if(radar_color >= 0) {
    // animate object radar reading
    putpixel(radar->frame,
      (int)(normal_x() / (X_MAX / 56.0)),
      (int)(normal_y() / (Y_MAX / 48.0)), pallete_color[radar_color]);
    radar->add_pixel(
      (int)(normal_x() / (X_MAX / 56.0)),
      (int)(normal_y() / (Y_MAX / 48.0)), 1, 1);
  }
}

void SpaceObject::collide(SpaceObject *other)
{
  double leta, kalfa;
  double alpha = 0.0;
  double beta  = 0.0;
  double v1, v2;
  double vk1, vk2, vl1, vl2;
  double ul1, ul2, uk1, uk2;
  double dx, dy;

  int x1, y1;
  int x2, y2;

  SpaceObject *big, *small;

  if((!canCollide(other)) || (!other->canCollide(this)))
    return;

  x1 = (int)(normal_x() - (w / 2.0));
  y1 = (int)(normal_y() - (h / 2.0));
  dx = min_delta(normal_x(), other->normal_x(), X_MAX);
  dy = min_delta(normal_y(), other->normal_y(), Y_MAX);
  x2 = (int)(normal_x() - dx - ((other->w) / 2.0));
  y2 = (int)(normal_y() - dy - ((other->h) / 2.0));

  if(!sprite->collide(x1, y1, sprite_index,
    x2, y2, other->sprite_index, other->sprite))
    return;

  if((mass == 0.0) || (other->mass == 0.0)) {
    inflict_damage(other);
    other->inflict_damage(this);
    return;
  }

  if((mass >= 0.0) && (other->mass >= 0.0)) {
    v1 = sqrt(((vx + driftvx) * (vx + driftvx)) +
              ((vy + driftvy) * (vy + driftvy)));
    v2 = sqrt(((other->vx + other->driftvx) * (other->vx + other->driftvx)) +
              ((other->vy + other->driftvy) * (other->vy + other->driftvy)));

    do {
      if(dx == 0.0)
        kalfa = 0.0;
      else
        kalfa = atan(dy / dx);
      if(dx < 0.0)
        kalfa += PI;
      leta = kalfa - PI;

      kalfa = normalize(kalfa, PI * 2.0);
      leta  = normalize(leta, PI * 2.0);

      x += (cos(kalfa) * v1);
      y += (sin(kalfa) * v1);
      other->x += (cos(leta) * v2);
      other->y += (sin(leta) * v2);

      x1 = (int)(normal_x() - (w / 2.0));
      y1 = (int)(normal_y() - (h / 2.0));
      dx = min_delta(normal_x(), other->normal_x(), X_MAX);
      dy = min_delta(normal_y(), other->normal_y(), Y_MAX);
      x2 = (int)(normal_x() - dx - ((other->w) / 2.0));
      y2 = (int)(normal_y() - dy - ((other->h) / 2.0));
    } while(sprite->collide(x1, y1, sprite_index,
      x2, y2, other->sprite_index, other->sprite));

    if(dx == 0.0)
      kalfa = 0.0;
    else
      kalfa = atan(dy / dx);
    if(dx < 0.0)
      kalfa += PI;
    leta = kalfa - (PI / 2.0);

    kalfa = normalize(kalfa, PI * 2.0);
    leta  = normalize(leta, PI * 2.0);

    if((vx + driftvx) == 0.0) {
      if((vy + driftvy) == 0.0)
        alpha = angle * ANGLE_RATIO;
      if((vy + driftvy) > 0.0)
        alpha = PI * 0.5;
      if((vy + driftvy) < 0.0)
        alpha = PI * 1.5;
    } else {
      alpha = atan((vy + driftvy) / (vx + driftvx));
      if((vx + driftvx) < 0.0)
        alpha += PI;
    }

    if((other->vx + other->driftvx) == 0.0) {
      if((other->vy + other->driftvy) == 0.0)
        beta = other->angle * ANGLE_RATIO;
      if((other->vy + other->driftvy) > 0.0)
        beta = PI * 0.5;
      if((other->vy + other->driftvy) < 0.0)
        beta = PI * 1.5;
    } else {
      beta = atan((other->vy + other->driftvy) / (other->vx + other->driftvx));
      if((other->vx + other->driftvx) < 0.0)
        beta += PI;
    }

    alpha = normalize(alpha, PI * 2.0);
    beta  = normalize(beta, PI * 2.0);

    vk1 = cos(kalfa - alpha) * v1;
    vk2 = cos(kalfa - beta)  * v2;
    vl1 = cos(leta - alpha) * v1;
    vl2 = cos(leta - beta)  * v2;

    ul1 = vl1;
    ul2 = vl2;

    uk1 = (2 * vk2 * other->mass + vk1 * mass - vk1 * other->mass);
    if((mass + other->mass) != 0.0)
      uk1 = uk1 / (mass + other->mass);
    uk2 = vk1 - vk2 + uk1;

    vx      = 0.0;
    vy      = 0.0;
    driftvx = cos(kalfa) * uk1 + cos(leta) * ul1;
    driftvy = sin(kalfa) * uk1 + sin(leta) * ul1;

    other->vx      = 0.0;
    other->vy      = 0.0;
    other->driftvx = cos(kalfa) * uk2 + cos(leta) * ul2;
    other->driftvy = sin(kalfa) * uk2 + sin(leta) * ul2;
  } else
  if((mass >= 0.0) || (other->mass >= 0.0)) {
    if(mass < 0.0) {
      small = other;
      big = this;
    } else {
      small = this;
      big = other;
    }

    v1 = sqrt(((small->vx + small->driftvx) * (small->vx + small->driftvx)) +
              ((small->vy + small->driftvy) * (small->vy + small->driftvy)));

    do {
      if(small->normal_x() == big->normal_x())
        leta = 0.0;
      else
        leta = atan(
          (small->normal_y() - big->normal_y()) /
          (small->normal_x() - big->normal_x()));
      if(small->normal_x() < big->normal_x())
        leta += PI;
      leta = normalize(leta, PI * 2.0);
      small->x += (cos(leta) * v1);
      small->y += (sin(leta) * v1);
    } while(sprite->collide(
      (int)(normal_x() - (w / 2.0)),
      (int)(normal_y() - (h / 2.0)), sprite_index,
      (int)(other->normal_x() - ((other->w) / 2.0)),
      (int)(other->normal_y() - ((other->h) / 2.0)),
      other->sprite_index, other->sprite));

    if((small->vx + small->driftvx) != 0.0) {
      alpha = atan((small->vy + small->driftvy) / (small->vx + small->driftvx));
      if((small->vx + small->driftvx) < 0.0)
        alpha += PI;
      if(small->normal_x() == big->normal_x())
        leta = 0.0;
      else
        leta = atan(
          (small->normal_y() - big->normal_y()) /
          (small->normal_x() - big->normal_x()));
      if(small->normal_x() < big->normal_x())
        leta += PI;

      alpha = normalize(alpha, PI * 2.0);
      leta  = normalize(leta, PI * 2.0);
      beta  = ((2.0 * leta) - alpha) + PI;

      small->vx      = 0.0;
      small->vy      = 0.0;
      small->driftvx = cos(beta) * v1;
      small->driftvy = sin(beta) * v1;
    }
  }

  inflict_damage(other);
  other->inflict_damage(this);
}

double SpaceObject::collide_line(double lx1, double ly1, double lx2,
  double ly2, double llength)
{
  int collide_x = (int)(lx2);
  int collide_y = (int)(ly2);
  double dx, dy;

  if(sprite->collide_line(
    (int)(lx1), (int)(ly1), &collide_x, &collide_y,
    (int)(lx1 - min_delta(lx1, normal_x(), X_MAX)),
    (int)(ly1 - min_delta(ly1, normal_y(), Y_MAX)), sprite_index)) {
    dx = lx2 - (double)(collide_x);
    dy = ly2 - (double)(collide_y);
    llength = llength - sqrt((dx * dx) + (dy * dy));
  }

  return(llength);
}

void SpaceObject::inflict_damage(SpaceObject *other)
{
  int i;

  if(damage_factor > 0) {
    i = damage_factor / 2;
    if(i >= BOOM_SAMPLES)
      i = BOOM_SAMPLES - 1;

    play_sample((SAMPLE *)(melee[MELEE_BOOM + i].dat), 255, 128, 1000, FALSE);

    other->damage += damage_factor;
  }

  other->handle_damage();
}

inline void SpaceObject::handle_damage() {}

void SpaceObject::calculate()
{
  x += (vx + driftvx);
  while(x < (double)(SCREEN_W - 64))
    x += (X_MAX * 4.0);
  while(x >= ((X_MAX * 4.0) + (double)(SCREEN_W - 64)))
    x -= (X_MAX * 4.0);

  y += (vy + driftvy);
  while(y < (double)(SCREEN_H))
    y += (Y_MAX * 4.0);
  while(y >= ((Y_MAX * 4.0) + (double)(SCREEN_H)))
    y -= (Y_MAX * 4.0);
}

void SpaceObject::pull(SpaceObject *other, double pullVelocity)
{
  driftvx += (cos((trajectory_angle(other)) * ANGLE_RATIO) *
    pullVelocity);
  driftvy += (sin((trajectory_angle(other)) * ANGLE_RATIO) *
    pullVelocity);
}

SpaceLine::SpaceLine(double lx, double ly, double langle, int lcflag,
  double llength, int lcolor) :
  SpaceLocation(lx, ly, langle, lcflag),
  length(llength),
  color(lcolor),
  collide_object(NULL),
  next(NULL)
{
  id = SPACE_LINE;
}

int SpaceLine::canCollide(SpaceLocation *other)
{
  if(!SpaceLocation::canCollide(other))
    return(FALSE);

  if(!(other->getID() & SPACE_OBJECT))
    return(FALSE);

  return(TRUE);
}

inline double SpaceLine::edge_x()
{
  return(cos(angle * ANGLE_RATIO) * length);
}

inline double SpaceLine::edge_y()
{
  return(sin(angle * ANGLE_RATIO) * length);
}

inline double SpaceLine::get_length()
{
  return(length);
}

void SpaceLine::calculate()
{
  int i;

  if((damage_factor > 0) && (collide_object)) {
    i = damage_factor / 2;
    if(i >= BOOM_SAMPLES)
      i = BOOM_SAMPLES - 1;
    play_sample((SAMPLE *)(melee[MELEE_BOOM + i].dat), 255, 128, 1000, FALSE);
    collide_object->damage += damage_factor;
    collide_object->handle_damage();
    damage_factor = 0;
    spaceFrame->addSpaceObject(LAYER_EXPLOSIONS, new Animation(
      normal_x() + edge_x(), normal_y() + edge_y(), sparkSprite,
      SPARK_FRAMES, 5));
  }

  if(collide_object)
    collide_object = NULL;
}

void SpaceLine::animate(Frame *space)
{
  double x1, y1, x2, y2;

  x1 = normal_x() - space_x;
  x1 = normalize(x1, X_MAX);
  x2 = x1 + edge_x();
  x1 *= space_zoom;
  x2 *= space_zoom;

  y1 = normal_y() - space_y;
  y1 = normalize(y1, Y_MAX);
  y2 = y1 + edge_y();
  y1 *= space_zoom;
  y2 *= space_zoom;

  line(space->frame, (int)(x1), (int)(y1), (int)(x2), (int)(y2), color);
  space->add_line((int)(x1), (int)(y1), (int)(x2), (int)(y2));
}

void SpaceLine::collide(SpaceObject *o)
{
  double old_length = length;

  if((!canCollide(o)) || (!o->canCollide(this)))
    return;

  length = o->collide_line(normal_x(), normal_y(), normal_x() + edge_x(),
    normal_y() + edge_y(), length);
  if(length != old_length)
    collide_object = o;
}

SpaceFrame::SpaceFrame(Frame *spaceFrame, Frame *radarFrame, int layerCount) :
  firstLine(NULL),
  space(spaceFrame),
  radar(radarFrame),
  numLayers(layerCount)
{
  int i;

  firstLayerObject = new (SpaceObject *)[numLayers];

  for(i = 0; i < numLayers; i++)
    firstLayerObject[i] = NULL;
}

SpaceFrame::~SpaceFrame()
{
  SpaceObject *ocurrent, *otemp;
  SpaceLine *lcurrent, *ltemp;
  int i;

  lcurrent = firstLine;
  while(lcurrent != NULL) {
    ltemp = lcurrent;
    lcurrent = lcurrent->next;
    delete ltemp;
  }

  for(i = 0; i < numLayers; i++) {
    ocurrent = firstLayerObject[i];
    while(ocurrent != NULL) {
      otemp = ocurrent;
      ocurrent = ocurrent->next;
      delete otemp;
    }
  }
}

SpaceObject *SpaceFrame::addSpaceObject(int layer, SpaceObject *o)
{
  o->next = firstLayerObject[layer];
  firstLayerObject[layer] = o;
  return(o);
}

SpaceLine *SpaceFrame::addSpaceLine(SpaceLine *l)
{
  l->next = firstLine;
  firstLine = l;
  return(l);
}

SpaceObject *SpaceFrame::getFirstObject(int layer)
{
  return firstLayerObject[layer];
}

SpaceLine *SpaceFrame::getFirstLine()
{
  return firstLine;
}

void SpaceFrame::calculate()
{
  SpaceObject *ocurrent, *otemp, *oprev;
  SpaceLine *lcurrent, *ltemp, *lprev;
  int i;

  for(i = 0; i < numLayers; i++) {
    ocurrent = firstLayerObject[i];
    while(ocurrent) {
      if(ocurrent->exists())
        ocurrent->calculate();
      ocurrent = ocurrent->next;
    }

    ocurrent = firstLayerObject[i];
    oprev = NULL;
    while(ocurrent) {
      if(!ocurrent->exists()) {
        otemp = ocurrent;
        ocurrent = ocurrent->next;
        if(oprev)
          oprev->next = ocurrent;
        if(firstLayerObject[i] == otemp)
          firstLayerObject[i] = ocurrent;
        delete otemp;
      } else {
        oprev = ocurrent;
        ocurrent = ocurrent->next;
      }
    }
  }

  lcurrent = firstLine;
  lprev = NULL;

  while(lcurrent) {
    if(lcurrent->exists())
      lcurrent->calculate();
    lcurrent = lcurrent->next;
  }

  lcurrent = firstLine;
  lprev    = NULL;

  while(lcurrent) {
    if(!lcurrent->exists()) {
      ltemp = lcurrent;
      lcurrent = lcurrent->next;
      if(lprev)
        lprev->next = lcurrent;
      if(firstLine == ltemp)
        firstLine = lcurrent;
      delete ltemp;
    } else {
      lprev = lcurrent;
      lcurrent = lcurrent->next;
    }
  }
}

void SpaceFrame::animate()
{
  SpaceObject *ocurrent;
  SpaceLine *lcurrent;
  int i;

  for(i = 0; i < numLayers; i++) {
    ocurrent = firstLayerObject[i];
    while(ocurrent != NULL) {
      if(ocurrent->exists())
        ocurrent->animate(space, radar);
      ocurrent = ocurrent->next;
    }
  }

  lcurrent = firstLine;
  while(lcurrent != NULL) {
    if(lcurrent->exists())
      lcurrent->animate(space);
    lcurrent = lcurrent->next;
  }
}

void SpaceFrame::collide()
{
  SpaceObject *ocurrent, *oother;
  SpaceLine *lcurrent, *lother;
  int i, j;

  for(i = 0; i < numLayers; i++) {
    ocurrent = firstLayerObject[i];
    while(ocurrent != NULL) {
      if(ocurrent->exists()) {
        oother = ocurrent->next;
        for(j = i; j < numLayers; j++) {
          if(i < j)
            oother = firstLayerObject[j];
          while(oother != NULL) {
            if(oother->exists())
              ocurrent->collide(oother);
            oother = oother->next;
          }
        }
      }
      ocurrent = ocurrent->next;
    }
  }

  for(i = 0; i < numLayers; i++) {
    lcurrent = firstLine;
    while(lcurrent != NULL) {
      if(lcurrent->exists()) {
        oother = firstLayerObject[i];
        while(oother != NULL) {
          if(oother->exists())
            lcurrent->collide(oother);
          oother = oother->next;
        }
      }
      lcurrent = lcurrent->next;
    }
  }
}