
#include <math.h>
#include "../melee.h"

// Define shot spark
SpaceSprite *sparkSprite;

Shot::Shot(double ox, double oy, double oangle, double ov, int odamage,
  double orange, int oarmour, Ship *oship, SpaceLocation *opos,
  SpaceSprite *osprite) :
  SpaceObject(opos->normal_x(), opos->normal_y(), oangle,
    oship->collideFlag(), osprite),
  ship(oship),
  v(ov),
  d(0.0),
  range(orange),
  armour(oarmour),
  explosionSprite(sparkSprite),
  explosionSample(NULL),
  explosionFrameCount(SPARK_FRAMES),
  explosionFrameSize(5)
{
  double alpha = (opos->get_angle() * ANGLE_RATIO) + (PI / 2.0);

  if(oy != 0.0)
    alpha += atan(oy / ox);
  if(ox < 0.0)
    alpha += PI;

  x += (sqrt((ox * ox) + (oy * oy)) * cos(alpha));
  y += (sqrt((ox * ox) + (oy * oy)) * sin(alpha));
  vx = (v * cos(angle * ANGLE_RATIO)) + opos->get_vx();
  vy = (v * sin(angle * ANGLE_RATIO)) + opos->get_vy();

  damage_factor = odamage;

  id = SPACE_SHOT;

  if(!ship->exists())
    state = 0;
}

void Shot::calculate()
{
  if(!ship->exists()) {
    state = 0;
    return;
  }

  SpaceObject::calculate();

  if(range < 0.0)
    return;
  if(d >= range)
    state = 0;
  d += v;
}

void Shot::handle_damage()
{
  if((damage > 0) || (directDamage > 0)) {
    armour -= damage;
    armour -= directDamage;
    damage = 0;
    directDamage = 0;
    if(armour <= 0) {
      armour = 0;
      state = 0;
      animateExplosion();
      soundExplosion();
    }
  }
}

void Shot::inflict_damage(SpaceObject *other)
{
  SpaceObject::inflict_damage(other);
  if (other->mass > 0.0) state = 0;
  animateExplosion();
}

void Shot::animateExplosion()
{
  spaceFrame->addSpaceObject(LAYER_EXPLOSIONS, new Animation(normal_x(),
    normal_y(), explosionSprite, explosionFrameCount, explosionFrameSize));
}

void Shot::soundExplosion()
{
  if(explosionSample) {
    stop_sample(explosionSample);
    play_sample(explosionSample, 255, 128, 1000, FALSE);
  }
}

void Shot::stop() {
  vx = 0.0;
  vy = 0.0;
  driftvx = 0.0;
  driftvy = 0.0;
}

void Shot::destroy() {
  state = 0;
}

AnimatedShot::AnimatedShot(double ox, double oy, double oangle, double ov,
  int odamage, double orange, int oarmour, Ship *oship, SpaceLocation *opos,
  SpaceSprite *osprite, int ofcount, int ofsize) :
  Shot(ox, oy, oangle, ov, odamage, orange, oarmour, oship, opos, osprite),
  frame_count(ofcount),
  frame_size(ofsize),
  frame_step(0)
{
}

void AnimatedShot::calculate()
{
  Shot::calculate();

  frame_step++;
  if(frame_step == frame_size) {
    frame_step = 0;
    sprite_index++;
    if(sprite_index == frame_count)
      sprite_index = 0;
  }
}

Missile::Missile(double ox, double oy, double oangle, double ov, int odamage,
  double orange, int oarmour, Ship *oship, SpaceLocation *opos,
  SpaceSprite *osprite) :
  Shot(ox, oy, oangle, ov, odamage, orange, oarmour, oship, opos, osprite)
{
  sprite_index = (int)(angle / 5.625) + 16;
  while(sprite_index >= 64)
    sprite_index -= 64;
  while(sprite_index < 0)
    sprite_index += 64;
}

HomingMissile::HomingMissile(double ox, double oy, double oangle, double ov,
  int odamage, double orange, int oarmour, double otrate, Ship *oship,
  SpaceSprite *osprite) :
  Missile(ox, oy, oangle, ov, odamage, orange, oarmour, oship, oship, osprite),
  turn_rate(20.0 / ((otrate + 1.0) * 5.0)),
  turn_step(0.0)
{
}

void HomingMissile::calculate()
{
  double alpha = angle;

  Missile::calculate();

  if((ship->target) && (!ship->target->isAlly(ship))) {
    alpha = trajectory_angle(ship->target);

    if(((alpha - angle) <= 180.0) && ((alpha - angle) >= 0.0) ||
       ((alpha - angle) <= (-180.0)))
      turn_step += turn_rate;
    else
      turn_step -= turn_rate;
  }

  while(fabs(turn_step) > 5.625) {
    if(turn_step < 0.0) {
      angle -= 5.625;
      turn_step += 5.625;
    } else
    if(turn_step > 0.0) {
      angle += 5.625;
      turn_step -= 5.625;
    }

    if(angle < 0.0)
      angle += 360.0;
    if(angle >= 360.0)
      angle -= 360.0;
  }

  sprite_index = (int)(angle / 5.625) + 16;
  if(sprite_index < 0)
    sprite_index += 64;
  if(sprite_index >= 64)
    sprite_index -= 64;

  vx = v * cos(angle * ANGLE_RATIO);
  vy = v * sin(angle * ANGLE_RATIO);
}

Laser::Laser(double langle, int lcolor, double lrange, int ldamage,
  int lfcount, SpaceLocation *opos, double rel_x, double rel_y) :
  SpaceLine(0.0, 0.0, langle, opos->collideFlag(), lrange, lcolor),
  frame(0),
  frame_count(lfcount),
  pos(opos),
  pos_x(rel_x),
  pos_y(rel_y)
{
  double alpha;

  id            = SPACE_LASER;
  damage_factor = ldamage;

  alpha = (pos->get_angle() * ANGLE_RATIO) + (PI / 2.0);
  if(pos_y != 0.0)
    alpha += atan(pos_y / pos_x);
  if(pos_x < 0.0)
    alpha += PI;

  x = pos->normal_x() +
    (sqrt((pos_x * pos_x) + (pos_y * pos_y)) * cos(alpha));
  y = pos->normal_y() +
    (sqrt((pos_x * pos_x) + (pos_y * pos_y)) * sin(alpha));

  if(!pos->exists())
    state = 0;
}

void Laser::calculate()
{
  double alpha;

  if((frame < frame_count) && (pos->exists())) {
    alpha = (pos->get_angle() * ANGLE_RATIO) + (PI / 2.0);
    if(pos_y != 0.0)
      alpha += atan(pos_y / pos_x);
    if(pos_x < 0.0)
      alpha += PI;

    x = pos->normal_x() +
      (sqrt((pos_x * pos_x) + (pos_y * pos_y)) * cos(alpha));
    y = pos->normal_y() +
      (sqrt((pos_x * pos_x) + (pos_y * pos_y)) * sin(alpha));
    SpaceLine::calculate();
    frame++;
  } else
    state = 0;
}