
#include <math.h>
#include "../melee.h"

class AndrosynthGuardian : public Ship {
  double       weaponRange;
  double       weaponVelocity;
  int          weaponDamage;
  int          weaponArmour;

  SpaceSprite *specialSprite;
  SpaceSprite *shipSprite;
  double       specialVelocity;
  double       specialTurnRate;
  int          specialDamage;
  int          specialActive;
  double       shipTurnRate;

  public:
  AndrosynthGuardian(double ox, double oy, double shipAngle,
    int shipCollideFlag, int shipRadarColor, ShipData *shipData);

  virtual void calculate();
  virtual void calculate_thrust();
  virtual void calculate_hotspots();

  virtual void inflict_damage(SpaceObject *other);

  virtual int activate_weapon();
  virtual int activate_special();
};

class AndrosynthBubble : public AnimatedShot {
  int courseFrames;

  public:
  AndrosynthBubble(double ox, double oy, double oangle, double ov,
    int odamage, double orange, int oarmour, Ship *oship, SpaceSprite *osprite,
    int ofcount, int ofsize);

  void calculate();
};

AndrosynthGuardian::AndrosynthGuardian(double ox, double oy, double shipAngle,
  int shipCollideFlag, int shipRadarColor, ShipData *shipData) :
  Ship(ox, oy, shipAngle, shipCollideFlag, shipRadarColor, shipData)
{
  set_config_data((char *)(data->data[0].dat), data->data[0].size);

  weaponRange        = (double)(get_config_int("Weapon", "Range", 0)) * RANGE_RATIO;
  weaponVelocity     = (double)(get_config_int("Weapon", "Velocity", 0)) / VELOCITY_RATIO;
  weaponDamage       = get_config_int("Weapon", "Damage", 0);
  weaponArmour       = get_config_int("Weapon", "Armour", 0);

  specialSprite   = data->spriteSpecial;
  shipSprite      = sprite;
  specialDamage   = get_config_int("Special", "Damage", 0);
  specialVelocity = (double)(get_config_int("Special", "Velocity", 0)) / VELOCITY_RATIO;
  specialTurnRate = (double)(get_config_int("Special", "TurnRate", 0));
  specialActive   = FALSE;

  specialTurnRate = 20.0 / ((specialTurnRate + 1.0) * 5.0);
  shipTurnRate    = turn_rate;
}

void AndrosynthGuardian::calculate()
{
  if(specialActive) {
    turn_rate = specialTurnRate;
    if(batt == 0) {
      sprite = shipSprite;
      w      = shipSprite->width();
      h      = shipSprite->height();
      damage_factor = 0;
      specialActive = FALSE;
      vx = 0.0;
      vy = 0.0;
      turn_rate = shipTurnRate;
    } else
    if(special_recharge == 0) {
      batt -= 2;
      special_recharge = recharge_rate;
    }
  }

  Ship::calculate();
}

void AndrosynthGuardian::calculate_thrust()
{
  if(specialActive) {
    vx = specialVelocity * cos(angle * ANGLE_RATIO);
    vy = specialVelocity * sin(angle * ANGLE_RATIO);
    driftvx *= 0.9;
    driftvy *= 0.9;
  } else
    Ship::calculate_thrust();
}

void AndrosynthGuardian::calculate_hotspots()
{
  if(specialActive)
    return;
  Ship::calculate_hotspots();
}

void AndrosynthGuardian::inflict_damage(SpaceObject *other)
{
  if(damage_factor > 0)
    spaceFrame->addSpaceObject(LAYER_EXPLOSIONS, new Animation(
      x + (cos(trajectory_angle(other) * ANGLE_RATIO) * 20.0),
      y + (sin(trajectory_angle(other) * ANGLE_RATIO) * 20.0),
      sparkSprite, SPARK_FRAMES, 5));
  SpaceObject::inflict_damage(other);
}

int AndrosynthGuardian::activate_weapon() {
  spaceFrame->addSpaceObject(LAYER_SHOTS, new AndrosynthBubble(
    0.0, -(height() / 2.0), angle, weaponVelocity, weaponDamage, weaponRange,
    weaponArmour, this, data->spriteWeapon, 10, 5));
  return(TRUE);
}

int AndrosynthGuardian::activate_special() {
  if(specialActive)
    return(FALSE);

  sprite = specialSprite;
  w      = specialSprite->width();
  h      = specialSprite->height();

  damage_factor = specialDamage;
  specialActive = TRUE;
  shipTurnRate  = turn_rate;
  turn_rate     = specialTurnRate;

  return(TRUE);
}

AndrosynthBubble::AndrosynthBubble(double ox, double oy, double oangle,
  double ov, int odamage, double orange, int oarmour, Ship *oship,
  SpaceSprite *osprite, int ofcount, int ofsize) :
  AnimatedShot(ox, oy, oangle, ov, odamage, orange, oarmour, oship, oship,
    osprite, ofcount, ofsize),
  courseFrames(10)
{
  explosionSprite     = ship->data->spriteWeaponExplosion;
  explosionFrameCount = 6;
  explosionFrameSize  = 5;
}

void AndrosynthBubble::calculate()
{
  double newAngle;

  AnimatedShot::calculate();

  courseFrames--;
  if(courseFrames == 0) {
    courseFrames = 10;
    newAngle = trajectory_angle(ship->target);
    newAngle -= 90.0;
    newAngle += (double)(random() % 180);
    vx = v * cos(newAngle * ANGLE_RATIO);
    vy = v * sin(newAngle * ANGLE_RATIO);
  }
}

Ship *getAndrosynthGuardian(double sX, double sY, double sAngle,
  int sCollideFlag, int sRadarColor, ShipData *sData)
{
  return(new AndrosynthGuardian(
    sX, sY, sAngle, sCollideFlag, sRadarColor, sData));
}

SHIP_REGISTER_INFO AndrosynthGuardianInfo =
{
  "andgu",
  "Androsynth Guardian",
  getAndrosynthGuardian
};
